"""MS SQL Server database"""
import pymssql
from ..base import Base
from ..dialect import Dialect


class DialectMssql(Dialect):
    """Mssql dialect"""

    @classmethod
    def limit(cls, limit, offset=None):
        """
        Require SQL Server >= 2012
        """
        if offset:
            return (
                "OFFSET {} ROWS FETCH NEXT {} ROWS ONLY".format(offset, limit),
                -1,
            )
        return "TOP {}".format(limit), 1


class Mssql(Base):
    """Mssql medoo wrapper"""

    def __init__(self, *args, **kwargs):
        super(Mssql, self).__init__(*args, **kwargs)
        self.cursor = self.connection.cursor()
        self.dialect(DialectMssql)

    def _connect(self, *args, **kwargs):
        arguments = {"server": ".", "port": 1433}
        arguments.update(kwargs)
        return pymssql.connect(**arguments)



"""Mysql database"""
import mysql.connector
from ..base import Base
from ..dialect import Dialect


class _MysqlConnectorCursor:
    """Wrap up mysql.connector.cursor object
    When there is no more records, mysql.connector.cursor returns None
    However, more pythonic way is to raise a StopIteration Exception.
    """

    def __init__(self, mccursor):
        self._mccursor = mccursor

    def __getattr__(self, name):
        return getattr(self._mccursor, name)

    def __next__(self):
        ret = next(self._mccursor)
        if ret is None:
            raise StopIteration()
        return ret

    def __iter__(self):
        return self


class DialectMysql(Dialect):
    """Mysql dialect"""

    @staticmethod
    def quote(item):
        if isinstance(item, str):
            if item == "*":
                return item
            return "`%s`" % item.replace("`", "``")
        return str(item)

    @staticmethod
    def value(item):
        if isinstance(item, str):
            # borrowed from
            # https://github.com/PyMySQL/PyMySQL/blob/3e71dd32e8ce868b090c282759eebdeabc960f58/pymysql/converters.py#L64
            # fixes #8
            _escape_table = [chr(x) for x in range(128)]
            _escape_table[0] = "\\0"
            _escape_table[ord("\\")] = "\\\\"
            _escape_table[ord("\n")] = "\\n"
            _escape_table[ord("\r")] = "\\r"
            _escape_table[ord("\032")] = "\\Z"
            _escape_table[ord('"')] = '\\"'
            _escape_table[ord("'")] = "\\'"
            return "'%s'" % item.translate(_escape_table)
        return str(item)


class Mysql(Base):
    """Mysql medoo wrapper"""

    def __init__(self, *args, **kwargs):
        super(Mysql, self).__init__(*args, **kwargs)
        self.cursor = _MysqlConnectorCursor(
            self.connection.cursor(buffered=True)
        )
        self.dialect(DialectMysql)

    def _connect(self, *args, **kwargs):
        arguments = {"host": "localhost", "port": 3306}
        arguments.update(kwargs)
        return mysql.connector.connect(**arguments)



"""The base for pymedoo"""

from .builder import Builder
from .record import Records
from .dialect import Dialect


class Base:
    """The base class medoo"""

    def __init__(self, *args, **kwargs):
        if "logging" in kwargs:
            self.logging = kwargs["logging"]
            del kwargs["logging"]
        else:
            self.logging = False

        self._dialect = None
        if "dialect" in kwargs:
            self._dialect = kwargs["dialect"]
            self.dialect(kwargs["dialect"])
            del kwargs["dialect"]

        # in case self._connect raises error
        self.connection = None
        self.connection = self._connect(*args, **kwargs)
        self.cursor = self.connection.cursor()
        self.history = []
        self.errors = []
        self.sql = None

    @property
    def builder(self):
        """Get the builder"""
        return Builder(dialect=self._dialect)

    def id(self):  # pylint: disable=invalid-name
        """Get the last insert id"""
        return self.cursor.lastrowid

    def _connect(self, *args, **kwargs):
        raise NotImplementedError("API not implemented.")

    def close(self):
        """Close the connection"""
        self.connection.close()

    def dialect(self, dial=None):
        """Set the dialect"""
        dial = dial or Dialect
        self._dialect = dial
        Builder.DIALECT = dial

    def __del__(self):
        if self.connection:
            self.close()

    def last(self):
        """Return the last action"""
        return self.history[-1] if self.history else ""

    def log(self):
        """Return the history"""
        return self.history

    def error(self):
        """Return the errros"""
        return self.errors

    def commit(self):
        """Commit the changes"""
        try:
            self.connection.commit()
        except Exception as ex:
            self.connection.rollback()
            raise ex

    # If data is an ordered dict, then datas could be tuples
    # otherwise, datas also should be dicts
    def insert(self, table, fields, *values, **kwargs):
        """INSERT clause"""
        sql = self.builder.insert(table, fields, *values)
        return self.query(sql, kwargs.get("commit", True))

    def update(self, table, data, where=None, commit=True):
        """UPDATE clause"""
        sql = self.builder.update(table, data, where)
        return self.query(sql, commit=commit)

    # where required to avoid all data deletion
    def delete(self, table, where, commit=True):
        """DELETE clause"""
        sql = self.builder.delete(table, where)
        return self.query(sql, commit=commit)

    def select(
        self,
        table,
        columns="*",
        where=None,
        join=None,
        distinct=False,
        newtable=None,
        sub=None,
        commit=False,
        readonly=True,
    ):
        """SELECT clause"""
        sql = self.builder.select(
            table, columns, where, join, distinct, newtable, sub
        )
        return self.query(sql, commit, readonly)

    def union(self, *queries, **kwargs):
        """Union statement"""
        sql = self.builder.union(*queries)
        return self.query(sql, commit=kwargs.get("commit", False))

    def has(self, table, where=None, join=None):
        """Has statement"""
        # pylint: disable=invalid-name
        rs = self.select(table, "*", where, join)
        return bool(rs.first())

    def get(self, table, columns="*", where=None, join=None):
        """Get a single value"""
        # pylint: disable=invalid-name
        rs = self.select(table, columns, where, join)
        return rs.first()[0]

    def query(self, sql, commit=True, readonly=True):
        """Send query to the connection"""
        self.sql = ("%s" % sql).strip()
        if self.logging:
            self.history.append(self.sql)
        else:
            self.history = [self.sql]
        try:
            self.cursor.execute(self.sql)
            if commit:
                self.commit()
            if self.sql.upper().startswith("SELECT"):
                return Records(self.cursor, readonly)
            return True
        except Exception as ex:
            self.errors.append(str(ex))
            if len(self.sql) <= 256:
                raise type(ex)(f"{ex}:\n{'-' * 32}\n{self.sql}")
            else:
                raise type(ex)(
                    f"{ex}:\n{'-' * 32}\n"
                    f"{self.sql[:256]} ...\n"
                    f"{'-' * 32}\n"
                    f"The above sql is slimed, full length: {len(self.sql)}"
                )



"""SQL Builder"""
import re
from .util import always_list
from .exception import (
    FieldParseError,
    TableParseError,
    WhereParseError,
    UpdateParseError,
    JoinParseError,
    LimitParseError,
    InsertParseError,
)
from .dialect import Dialect


class Term:
    """
    Any part of a sql statement could be a Term
    """

    def __str__(self):
        raise NotImplementedError(
            "API not implemented: " "cannot convert to string."
        )

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(str(self))


class Raw(Term):
    """
    Raw strings
    """

    # pylint: disable=invalid-name
    def __init__(self, s):
        self.s = str(s)

    def __str__(self):
        return self.s


class Table(Term):
    """
    Only table or schema.table
    """

    # a.c(alias)
    REGEX_FROM = r"^\s*((?:[\w_]+\.)?(?:[\w_]+|\*))\s*(?:\(([\w_]+)\))?\s*$"
    # a.c[=] # comment
    REGEX_JOIN = r"^\s*([\w_]+\.)?([\w_]+)\s*(?:\[(.+?)\])?\s*(?:#.*)?$"

    def __init__(self, table, schema=None):
        parts = table.split(".")
        if len(parts) == 1:
            self.table = table
            self.schema = schema
        elif len(parts) == 2:
            if schema:
                raise TableParseError(
                    "Confusing schemas specified in arguments "
                    '"table" and "schema"'
                )
            self.table = parts[1]
            self.schema = parts[0]
        else:
            raise TableParseError('Additional parts in "table"')

    def __str__(self):
        parts = []
        if self.schema:
            parts.append(Builder.DIALECT.quote(self.schema))
        parts.append(Builder.DIALECT.quote(self.table))
        return ".".join(parts)

    @staticmethod
    def parse(tablestr, context=None):
        """Parse a string of table as a Table object"""
        if isinstance(tablestr, Term):
            return tablestr

        if not context:
            return Table(tablestr)
        if context == "from":
            matching = re.match(Table.REGEX_FROM, tablestr)
            if not matching:
                raise TableParseError("Unrecognized table string.")
            table = matching.group(1)
            alias = matching.group(2)
            return TableFrom(table, alias=alias)
        if context in ["insert", "update", "delete", "selectinto"]:
            return Table(tablestr)
        raise TableParseError("Unknown table context: {}".format(context))


class TableFrom(Term):
    """Tables in FROM as Terms"""

    def __init__(self, table, alias=None):
        self.table = Table(table)
        self.alias = alias

    def __str__(self):
        return str(self.table) + (
            " AS " + Builder.DIALECT.quote(self.alias) if self.alias else ""
        )


class Field(Term):
    """
    Only field or table.field or schema.table.field
    """

    # a.b.c(alias)
    REGEX_SELECT = (
        r"^\s*((?:[\w_]+\.)?(?:[\w_]+\.)?(?:[\w_]+|\*))"
        r"\s*(?:\|\s*(\.?[\w_]+))?\s*(?:\(([\w_]+)\))?\s*$"
    )
    # a.b.f1, a.b.f2(func1, func2)[=] # comment
    # REGEX_WHERE = r'^\s*([\w\s_.]+)\s*(?:\(([\w_,]+)\))?\s*
    # (?:\[(.+?)\])?\s*(?:#.*)?$'

    def __init__(self, field="*", table=None, schema=None):
        parts = field.split(".")
        if len(parts) == 1:
            self.field = field
            self.table = table
            self.schema = schema
        elif len(parts) == 2:
            if table:
                raise FieldParseError(
                    "Confusing table specified in "
                    'arguments "field" and "table"'
                )
            self.field = parts[1]
            self.table = parts[0]
            self.schema = schema
        elif len(parts) == 3:
            if table or schema:
                raise FieldParseError(
                    'Confusing table/schema specified in arguments "field" '
                    'and "table"/"schema"'
                )
            self.field = parts[2]
            self.table = parts[1]
            self.schema = parts[0]
        else:
            raise FieldParseError('Additional parts in "field"')

    def __str__(self):
        parts = []
        if self.schema:
            parts.append(Builder.DIALECT.quote(self.schema))
        if self.table:
            parts.append(Builder.DIALECT.quote(self.table))
        parts.append(Builder.DIALECT.quote(self.field))
        return ".".join(parts)

    def __add__(self, value):
        return Raw(str(self) + "+" + Builder.DIALECT.value(value))

    def __sub__(self, value):
        return Raw(str(self) + "-" + Builder.DIALECT.value(value))

    def __mul__(self, value):
        return Raw(str(self) + "*" + Builder.DIALECT.value(value))

    def __div__(self, value):
        return Raw(str(self) + "/" + Builder.DIALECT.value(value))

    def __mod__(self, value):
        return Raw(str(self) + "%" + Builder.DIALECT.value(value))

    @staticmethod
    def parse(fieldstr, context=None):
        """Parse the field string"""
        if isinstance(fieldstr, Term):
            return fieldstr

        if not context:
            return Field(fieldstr)
        if context == "select":
            matching = re.match(Field.REGEX_SELECT, fieldstr)
            if not matching:
                raise FieldParseError(
                    "Unrecognized field string: " "{}".format(fieldstr)
                )
            field = matching.group(1)
            func = matching.group(2)
            alias = matching.group(3)
            return FieldSelect(field, alias=alias, func=func)
        if context == "group":
            return Field(fieldstr)
        raise FieldParseError("Unknown field context: {}".format(context))


class FieldSelect(Term):
    """Terms in SELECT"""

    def __init__(self, field="*", alias=None, func=None):
        """
        The fields in SELECT substatements
        @params:
            `field`   : The field that applied
            `alias`    : The alias
        """
        self.field = Field(field)
        self.alias = alias
        self.distinct = func and func.startswith(".")
        self.func = func[1:] if func and func.startswith(".") else func

    def __str__(self):
        ret = str(self.field)
        if self.func:
            if hasattr(Builder.DIALECT, self.func.lower()):
                ret = getattr(Builder.DIALECT, self.func.lower())(
                    ret, distinct=self.distinct
                )
            else:
                ret = "{}({}{})".format(
                    self.func.upper(),
                    "DISTINCT " if self.distinct else "",
                    ret,
                )
        if self.alias:
            ret += " AS " + Builder.DIALECT.quote(self.alias)
        return ret


class Where(Term):
    """WHERE as a Term"""

    def __init__(self, conditions, root=True):
        self.conditions = conditions
        self.root = root

    def __str__(self):
        sqlitems = []
        for key, val in self.conditions.items():
            # whatever term it is, the value will be ignored
            if isinstance(key, Term):
                sqlitems.append(str(key))
            elif key.split("#")[0].strip().upper() == "AND":
                if not isinstance(val, (tuple, list, dict)):
                    raise WhereParseError(
                        "Expect dict or item list/tuple for "
                        'conditions to be connected by AND: "%s"' % val
                    )
                val = (
                    val.items()
                    if isinstance(val, dict)
                    else [
                        v if isinstance(v, tuple) else (v, None) for v in val
                    ]
                )
                whereterms = [str(Where({k: v}, False)) for k, v in val]
                if len(whereterms) == 1:
                    sqlitems.append(whereterms[0])
                else:
                    sqlitems.append("(%s)" % (" AND ".join(whereterms)))
            elif key.split("#")[0].strip().upper() == "OR":
                if not isinstance(val, (tuple, list, dict)):
                    raise WhereParseError(
                        "Expect dict or item list for conditions to "
                        'be connected by OR: "%s"' % val
                    )
                val = (
                    val.items()
                    if isinstance(val, dict)
                    else [
                        v if isinstance(v, tuple) else (v, None) for v in val
                    ]
                )
                whereterms = [str(Where({k: v}, False)) for k, v in val]
                if len(whereterms) == 1:
                    sqlitems.append(whereterms[0])
                else:
                    sqlitems.append("(%s)" % (" OR ".join(whereterms)))
            else:
                sqlitems.append(str(WhereTerm(key, val)))

        if (
            self.root
            and len(sqlitems) == 1
            and sqlitems[0][0] == "("
            and sqlitems[0][-1] == ")"
        ):
            # remove the brackets
            return sqlitems[0][1:-1]
        return " AND ".join(sqlitems)


class WhereTerm(Term):
    """Terms in WHERE clause"""

    REGEX_KEY = (
        r"^\s*(!)?\s*([\w\s_.]+)\s*(?:\|([\w\s_.]+))?"
        r"\s*(?:\[(.+?)\])?\s*(?:#.*)?$"
    )

    def __init__(self, key, val):
        self.key = key
        self.val = val

    def __str__(self):
        # whatever term it is for the key, ignore the value
        if isinstance(self.key, Term):
            return str(self.key)

        matching = re.match(WhereTerm.REGEX_KEY, self.key)
        if not matching:
            raise WhereParseError("Unrecognized key in where conditions.")
        ret = "NOT " if matching.group(1) else ""
        field = FieldSelect(matching.group(2), func=matching.group(3))
        oprt = matching.group(4)

        return ret + Builder.DIALECT._operator(oprt, field, self.val)


class Order(Term):
    """ORDER as a term"""

    def __init__(self, orders):
        self.orders = orders

    def __str__(self):
        return ",".join(
            [str(OrderTerm(key, val)) for key, val in self.orders.items()]
        )


class OrderTerm(Term):
    """Terms in ORDER"""

    REGEX_KEY = (
        r"^\s*((?:[\w_]+\.)?(?:[\w_]+\.)?(?:[\w_]+|\*))"
        r"\s*(?:\|([\w_.]+))?\s*$"
    )

    def __init__(self, key, val):
        matching = re.match(OrderTerm.REGEX_KEY, key)
        if not matching:
            raise FieldParseError("Unrecognized field in ORDER BY clause.")
        self.field = FieldSelect(matching.group(1), func=matching.group(2))
        if val is True or val is None:
            val = "ASC"
        elif val is False:
            val = "DESC"
        else:
            val = val.upper()
        self.val = val

    def __str__(self):
        return "{} {}".format(self.field, self.val)


class Limit(Term):
    """Terms in Limit"""

    def __init__(self, limoff):
        if len(limoff) == 1:
            limit = limoff[0]
            offset = None
        elif len(limoff) == 2:
            limit, offset = limoff
        else:
            raise LimitParseError("LIMIT requires a two integer tuple/list.")

        lim = Builder.DIALECT.limit(limit, offset)
        if isinstance(lim, tuple):
            self.s, self.pos = lim  # pylint: disable=invalid-name
        else:
            self.s = lim  # pylint: disable=invalid-name
            self.pos = -1

    def __str__(self):
        return self.s


class Set(Term):
    """Set as a term"""

    def __init__(self, sets):
        self.sets = sets

    def __str__(self):
        return ",".join(
            [str(SetTerm(key, val)) for key, val in self.sets.items()]
        )


class SetTerm(Term):
    """Terms in SET"""

    REGEX_KEY = (
        r"^\s*((?:[\w_]+\.)?(?:[\w_]+\.)?"
        r"(?:[\w_]+|\*))\s*(?:\[(.+?)\])?\s*$"
    )

    def __init__(self, key, val):
        matching = re.match(SetTerm.REGEX_KEY, key)
        if not matching:
            raise UpdateParseError("Unrecognized field in UPDATE SET.")
        self.field = Field(matching.group(1))
        self.oprt = matching.group(2)
        self.val = val

    def __str__(self):
        return Builder.DIALECT._update(self.oprt, self.field, self.val)


class Join(Term):
    """JOIN as a term"""

    def __init__(self, joins, maintable=None):
        self.joins = joins
        self.maintable = maintable

    def __str__(self):
        return " ".join(
            [
                str(JoinTerm(key, val, self.maintable))
                for key, val in self.joins.items()
            ]
        )


class JoinTerm(Term):
    """Terms in JOIN clause"""

    REGEX_KEY = (
        r"^\s*(?:\[(.+?)\])?\s*((?:[\w_]+\.)?"
        r"(?:[\w_]+\.)?(?:[\w_]+|\*))\s*(?:\((.+?)\))?\s*$"
    )

    def __init__(
        self, key, val, maintable=None
    ):  # pylint: disable=too-many-branches
        matching = re.match(JoinTerm.REGEX_KEY, key)
        if not matching:
            raise JoinParseError("Unrecognized table in JOIN.")
        self.jointype = matching.group(1)
        self.table = TableFrom(matching.group(2), alias=matching.group(3))
        self.onfields = []
        fieldtable = self.table.alias or self.table.table
        if isinstance(maintable, TableFrom):
            maintable = maintable.alias or maintable.table
        elif isinstance(maintable, Builder):
            # subquery
            if maintable._subas in [None, True, False]:
                raise JoinParseError(
                    "Require alias for subquery to "
                    "refer to its fields in JOIN ON."
                )
            maintable = maintable._subas

        if isinstance(val, (tuple, list)):
            for iterval in val:
                field1 = Field(iterval)
                if field1.table:
                    raise JoinParseError(
                        "Expected table specified " "in JOIN ON fields."
                    )
                if not maintable:
                    raise JoinParseError(
                        "Short format of JOIN fields are "
                        "not allowed without primary table."
                    )
                field2 = Field(iterval)
                field1.table = fieldtable
                field2.table = maintable
                self.onfields.append((field1, field2))
        elif isinstance(val, dict):  # dict
            for iterkey, iterval in val.items():
                field1 = Field(iterkey)
                field2 = Field(iterval)

                if field1.table:
                    raise JoinParseError(
                        "Unexpected table on JOIN ON "
                        "left field: {}".format(field1)
                    )

                if maintable and field2.table:
                    raise JoinParseError(
                        "Unexpected table on JOIN ON "
                        "right field: {}".format(field2)
                    )

                field1.table = fieldtable
                if maintable:
                    field2.table = maintable
                self.onfields.append((field1, field2))
        else:
            field1 = Field(val)
            if field1.table:
                raise JoinParseError(
                    "Unexpected table specified in " "JOIN ON fields."
                )

            if not maintable:
                raise JoinParseError(
                    "Short format of JOIN fields are "
                    "not allowed without primary table."
                )

            field2 = Field(val)
            field1.table = fieldtable
            field2.table = maintable
            self.onfields.append((field1, field2))

    def __str__(self):
        return "{} {} ON {}".format(
            Builder.DIALECT._join(self.jointype),
            self.table,
            " AND ".join(
                "{}={}".format(key, val) for key, val in self.onfields
            ),
        )


class Builder(Term):
    """SQL builder"""

    DIALECT = Dialect

    def __init__(self, dialect=None):
        Builder.DIALECT = dialect or Dialect

        # for join
        self.table = None
        self.terms = []
        self._sql = None
        self._subas = False

    def _select(self, *fields, **kwargs):
        fields = fields or ["*"]
        distinct = kwargs.get("distinct", False)
        self.terms.append("SELECT DISTINCT" if distinct else "SELECT")

        fieldterms = [Field.parse(field or "*", "select") for field in fields]
        self.terms.append(",".join([str(field) for field in fieldterms]))
        return self

    def _sub(self, alias=None):
        """
        Give a name for subquery
        """
        self._subas = alias if alias else True
        return self

    def _from(self, *tables):
        self.terms.append("FROM")
        tableterms = [Table.parse(table, "from") for table in tables]

        self.table = (
            tableterms[0]._subas
            if isinstance(tableterms[0], Builder)
            else tableterms[0]
        )
        self.terms.append(",".join([str(table) for table in tableterms]))
        return self

    def _where(self, conditions):
        if not conditions:
            return self
        self.terms.append("WHERE")
        whereterm = Where(conditions)
        self.terms.append(whereterm)
        return self

    def _order(self, orders):
        self.terms.append("ORDER BY")
        self.terms.append(Order(orders))
        return self

    def _limit(self, limoff):
        if not isinstance(limoff, (tuple, list)):
            limoff = [limoff]
        lim = Limit(limoff)
        if lim.pos == -1:
            self.terms.append(lim)
        elif isinstance(lim.pos, int):
            self.terms.insert(lim.pos, str(lim))
        else:  # oracle, add where condition
            # check if where is there:
            if "WHERE" not in self.terms:
                self.terms.append("WHERE")
                self.terms.append(lim)
            else:
                whereindex = self.terms.index("WHERE")
                # TODO: ' AND ' and ' OR ' could also be in subquery or values
                # it's still OK to have brackets, keep it for now
                if " AND " in str(self.terms[whereindex + 1]) or " OR " in str(
                    self.terms[whereindex + 1]
                ):
                    self.terms[whereindex + 1] = "({}) AND ({})".format(
                        self.terms[whereindex + 1], lim
                    )
                else:
                    self.terms.insert(whereindex + 2, "AND ({})".format(lim))

        return self

    def _union(self, other, all_=False):
        self.terms.append("UNION")
        if all_:
            self.terms.append("ALL")
        self.terms.append(other)
        return self

    def _exists(self, query):
        self.terms.append("EXISTS")
        self.terms.append(query)
        return self

    def _group(self, fields):
        self.terms.append("GROUP BY")
        if isinstance(fields, Term):
            fields = [fields]
        else:
            fields = always_list(fields)
        fieldterms = [Field.parse(field, "group") for field in fields]
        self.terms.append(",".join([str(field) for field in fieldterms]))
        return self

    def _having(self, conditions):
        if not conditions:
            return self
        self.terms.append("HAVING")
        self.terms.append(Where(conditions))

        return self

    def _insert(
        self, table, values, fields=None
    ):  # pylint: disable=too-many-branches
        self.terms.append("INSERT INTO")
        self.terms.append(Table(table))
        if not fields:
            if isinstance(values[0], dict):
                fields = values[0].keys()
                for i, value in enumerate(values):
                    if i == 0:
                        continue
                    if isinstance(value, dict) and set(fields) != set(
                        value.keys()
                    ):
                        raise InsertParseError(
                            "Inconsistent keys in " "values for INSERT."
                        )
        if fields:
            fieldterms = [
                Table.parse(field, "insert") for field in always_list(fields)
            ]
            self.terms.append(
                "({})".format(",".join([str(field) for field in fieldterms]))
            )

        # support INSERT INTO SELECT ...
        if isinstance(values[0], Builder):
            subquery0 = values.pop(0)
            self.terms.append(subquery0)
            for value in values:
                if isinstance(value, Builder):
                    if value._subas:
                        self.terms.append("UNION ALL")
                        value._subas = None
                        self.terms.append(value)
                    else:
                        self.terms.append("UNION")
                        self.terms.append(value)
                else:
                    raise InsertParseError(
                        "All values should be subqueries " "in INSERT."
                    )
        else:
            insertvals = []
            for value in values:
                if isinstance(value, (tuple, list)):
                    insertvals.append(tuple(value))
                elif isinstance(value, dict):
                    if fields:
                        insertvals.append(
                            tuple(
                                value[field] if field in value else None
                                for field in fields
                            )
                        )
                    else:
                        insertvals.append(tuple(value.values()))
                else:
                    raise InsertParseError(
                        "Unsupported type for values "
                        "in INSERT: {}".format(type(value))
                    )
            self.terms.append("VALUES")
            valterms = []
            for inval in insertvals:
                valterms.append(
                    "(%s)"
                    % (",".join([Builder.DIALECT.value(iv) for iv in inval]))
                )
            self.terms.append(",".join(valterms))
        return self

    def _update(self, table):
        self.terms.append("UPDATE")
        self.terms.append(Table.parse(table, "update"))
        return self

    def _set(self, sets):
        self.terms.append("SET")
        self.terms.append(Set(sets))
        return self

    def _delete(self, table):
        self.terms.append("DELETE FROM")
        self.terms.append(Table.parse(table, "delete"))
        return self

    def _join(self, joins):
        self.terms.append(Join(joins, self.table))
        return self

    # support select * into newtable from ...
    def _into(self, table):
        self.terms.append("INTO")
        self.terms.append(Table.parse(table, "selectinto"))
        return self

    def select(
        self,
        table,
        columns="*",
        where=None,
        join=None,
        distinct=False,
        newtable=None,
        sub=None,
    ):
        """Build SELECT statement"""
        order = where.pop("ORDER", None) if where else None
        limit = where.pop("LIMIT", None) if where else None
        group = where.pop("GROUP", None) if where else None
        having = where.pop("HAVING", None) if where else None
        exists = where.pop("EXISTS", None) if where else None

        if isinstance(columns, Term):
            columns = [columns]
        elif columns is None:
            columns = "*"
        self._select(*always_list(columns), distinct=distinct)
        if newtable:
            self._into(newtable)._from(*always_list(table))
        else:
            self._from(*always_list(table))
        if join:
            self._join(join)._where(where)
        else:
            self._where(where)
        if order:
            self._order(order)
        if limit:
            self._limit(limit)
        if group:
            self._group(group)
        if having:
            self._having(having)
        if exists:
            self._exists(exists)
        if sub:
            self._sub(sub)
        return self

    def update(self, table, data, where=None):
        """Build UPDATE statement"""
        self._update(table)._set(data)._where(where)
        return self

    def delete(self, table, where):
        """Build DELETE statement"""
        self._delete(table)._where(where)
        return self

    def insert(self, table, fields, *values):
        """
        data: list will be treated as fields
        """
        # table, values, fields
        values2 = []
        if isinstance(fields, dict):
            values2.append(list(fields.values()))
            fields = list(fields.keys())
        elif isinstance(fields, tuple):
            values2.append(tuple(fields))
            fields = None
        elif isinstance(fields, list):
            pass
        else:  # assuming fields specified as string
            fields = always_list(fields)

        values2.extend(
            [
                value
                if isinstance(value, tuple)
                else tuple(value[key] for key in fields)
                for value in values
            ]
        )
        self._insert(table, values2, fields)
        return self

    def union(self, *queries):
        """Add UNION"""
        queries = list(queries)
        if not self.terms:
            query = queries.pop(0)
            subas = query._subas
            query._subas = None  # remove brackets
            self.terms.append(str(query))
            query._subas = subas
        for query in queries:
            if query._subas:
                subas = query._subas
                query._subas = None
                self.terms.append("UNION ALL")
                self.terms.append(str(query))
                query._subas = subas
            else:
                self.terms.append("UNION")
                self.terms.append(query)
        return self

    def sql(self):
        """Get the SQL"""
        if self._sql:
            return self._sql

        ret = " ".join(["%s" % t for t in self.terms])
        if self._subas is True:
            ret = "({})".format(ret)
        elif self._subas:
            ret = "({}) AS {}".format(ret, Builder.DIALECT.quote(self._subas))

        return ret

    def __str__(self):
        return self.sql()



"""Exceptions"""


class UpdateParseError(Exception):
    """Failed to parse update clause"""


class FieldParseError(Exception):
    """Failed to parse field"""


class TableParseError(Exception):
    """Failed to parse table"""


class WhereParseError(Exception):
    """Failed to parse where clause"""


class JoinParseError(Exception):
    """Failed to parse join clause"""


class AnyAllSomeParseError(Exception):
    """Failed to parse any, all or some modifiers"""


class LimitParseError(Exception):
    """Failed to parse limit"""


class InsertParseError(Exception):
    """Failed to parse insert clause"""


class RecordKeyError(KeyError):
    """KeyError for Record"""


class RecordAttributeError(AttributeError):
    """AttributeError for Record"""


class GetFromEmptyRecordError(Exception):
    """Try to get from empty record"""



"""A lightweight database framework for python"""

import importlib
from .builder import Raw, Table, Field
from .dialect import Dialect

__version__ = "0.1.2"

DATABASE_TYPES = {
    "Sqlite": ["sqlite", "sqlite3"],
    "Mysql": "mysql",
    "Mssql": "mssql",
    "Pgsql": ["pgsql", "postgres", "postgresql"],
}


class Medoo:
    """Main entrance"""

    def __new__(cls, dbtype, *args, **kwargs):

        for key, val in DATABASE_TYPES.items():
            if not isinstance(val, list):
                val = [val]
            if not dbtype.lower() in val:
                continue

            mod = importlib.import_module(
                ".database.{}".format(key.lower()), package="medoo"
            )
            klass = getattr(mod, key)
            return klass(*args, **kwargs)

        raise ValueError("Database type not supported: {}.".format(dbtype))



"""Sqlite3 adapter"""
import sqlite3
from ..base import Base
from ..dialect import Dialect


class DialectSqlite(Dialect):
    """Sqlite dialect"""

    @staticmethod
    def value(item):
        """Get the value"""
        if isinstance(item, str):
            return "'%s'" % item.replace("'", "''")
            # return "'{}'".format(item.replace("'", "''"))
        if isinstance(item, bool):
            return str(int(item))
        if item is None:
            return "NULL"
        return str(item)


class Sqlite(Base):
    """Sqlite medoo wrapper"""

    def __init__(self, *args, **kwargs):
        database = kwargs.pop("database", kwargs.pop("database_file", None))
        if database is not None and database.startswith("file://"):
            database = database.replace("file://", "")
        if database is not None:
            kwargs["database"] = database
        super(Sqlite, self).__init__(*args, **kwargs)
        self.cursor = self.connection.cursor()
        self.dialect(DialectSqlite)

    def _connect(self, *args, **kwargs):
        arguments = {
            "database": ":memory:",
            "timeout": 5.0,
            "detect_types": 0,
            "isolation_level": None,
            "check_same_thread": False,
            "cached_statements": 100,
            # 'factory'      : [str][0],
        }
        # 'database' had been made sure to replace 'database_file'
        # in __init__
        arguments.update(kwargs)
        return sqlite3.connect(**arguments)



"""Record fetched from database"""
from collections import OrderedDict
from .exception import (
    RecordKeyError,
    RecordAttributeError,
    GetFromEmptyRecordError,
)
from .util import reduce_datetimes


class Record:
    """
    A row, from a query, from a database.
    The idea is borrowed from https://github.com/kennethreitz/records
    """

    def __init__(self, keys, values, readonly=True):
        # sacrifice some efficiency but support setitem, setattr operation
        self.__dict__["_keys"] = keys if readonly else keys[:]
        self.__dict__["_values"] = values if readonly else values[:]
        self.__dict__["_readonly"] = readonly
        # Ensure that lengths match properly.
        assert len(self.keys()) == len(self.values())

    def keys(self):
        """
        Returns the list of column names from the query.
        """
        return self.__dict__["_keys"]

    def values(self):
        """
        Returns the list of values from the query.
        """
        return self.__dict__["_values"]

    def __repr__(self):
        return "<Record {}>".format(self.as_dict())

    def __getitem__(self, key):
        # Support for index-based lookup.
        if isinstance(key, int):
            try:
                return self.values()[key]
            except IndexError:
                raise GetFromEmptyRecordError("No records returned.")

        # Support for string-based lookup.
        keys = list(self.keys())
        if key in keys:
            if keys.count(key) > 1:
                raise RecordKeyError(
                    "Record contains multiple " "'{}' fields.".format(key)
                )
            return self.values()[keys.index(key)]

        raise RecordKeyError("Record contains no '{}' field.".format(key))

    def __setitem__(self, key, val):
        if self.__dict__["_readonly"]:
            raise RecordKeyError(
                "Readonly Record does not support " "setitem operation."
            )

        if isinstance(key, int):
            self.values()[key] = val
            return

        keycount = self.keys().count(key)
        if keycount > 1:
            raise RecordKeyError(
                "Record contains multiple " "'{}' fields.".format(key)
            )
        if keycount == 1:
            i = self.keys().index(key)
            self.values()[i] = val
        else:  # 0
            self.keys().append(key)
            self.values().append(val)

    def __delitem__(self, key):
        if self.__dict__["_readonly"]:
            raise RecordKeyError(
                "Readonly Record does not support " "delitem operation."
            )

        # Support for index-based lookup.
        if isinstance(key, int):
            del self.keys()[key]
            del self.values()[key]
            return

        # Support for string-based lookup.
        indexes = [i for i, k in enumerate(self.keys()) if k == key]
        for index in indexes:
            del self.keys()[index]
            del self.values()[index]
        if not indexes:
            raise RecordKeyError("Record contains no '{}' field.".format(key))

    def __getattr__(self, key):
        try:
            return self[key]
        except RecordKeyError as exc:
            raise RecordAttributeError(exc) from exc

    def __setattr__(self, key, val):
        if self.__dict__["_readonly"]:
            raise RecordAttributeError(
                "Readonly Record does not " "support setattr operation."
            )

        try:
            self[key] = val
        except RecordKeyError as exc:
            raise RecordAttributeError(exc) from exc

    def __dir__(self):
        standard = dir(super(Record, self))
        # Merge standard attrs with generated ones (from column names).
        return sorted(standard + [str(k) for k in self.keys()])

    def __eq__(self, other):
        if isinstance(other, OrderedDict):
            return self.as_dict(True) == other
        if isinstance(other, Record):
            return self.as_dict(True) == other.as_dict(True)
        return self.as_dict() == dict(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, key):
        return key in self.keys()

    def index(self, key):
        """Get the index of the key"""
        return self.keys().index(key)

    def get(self, key, default=None):
        """
        Returns the value for a given key, or default.
        """
        try:
            return self[key]
        except RecordKeyError:
            return default

    def items(self):
        """Get the items of the record"""
        for i, k in enumerate(self.keys()):
            yield k, self.values()[i]

    def as_dict(self, ordered=False):
        """
        Returns the row as a dictionary, as ordered.
        """
        items = zip(self.keys(), self.values())

        return OrderedDict(items) if ordered else dict(items)

    asDict = as_dict


class Records:
    """
    A set of excellent Records from a query.
    """

    def __init__(self, cursor, readonly=True):
        self.meta = [desc[0] for desc in cursor.description]
        self._cursor = cursor
        self._allrows = []
        self.pending = True
        self.readonly = readonly

    def __repr__(self):
        return "<Records: size={}, pending={}>".format(len(self), self.pending)

    def __iter__(self):
        """
        Iterate over all rows, consuming the underlying generator
        only when necessary.
        """
        i = 0
        while True:
            # Other code may have iterated between yields,
            # so always check the cache.
            if i < len(self):
                yield self[i]
            else:
                # Throws StopIteration when done.
                # Prevent StopIteration bubbling from generator,
                # following https://www.python.org/dev/peps/pep-0479/
                try:
                    yield next(self)
                except StopIteration:
                    return
            i += 1

    def __nonzero__(self):
        return self.first() is not None

    def __next__(self):
        try:
            nextrow = Record(
                self.meta, list(next(self._cursor)), readonly=self.readonly
            )
            self._allrows.append(nextrow)
            return nextrow
        except StopIteration:
            self.pending = False
            raise StopIteration("Records contains no more rows.")

    next = __next__

    def __getitem__(self, key):
        is_int = isinstance(key, int)

        # Convert RecordCollection[1] into slice.
        if is_int:
            key = slice(key, key + 1)

        while (key.stop is None) or (len(self) < key.stop):
            try:
                next(self)
            except StopIteration:
                break

        ret = self._allrows[key]
        return ret[0] if is_int else ret

    def __len__(self):
        return len(self._allrows)

    def export(self, format, **kwargs):  # pylint: disable=redefined-builtin
        """
        Export the RecordCollection to a given format (courtesy of Tablib).
        """
        return self.tldata.export(format, **kwargs)

    @property
    def tldata(self):
        """
        A Tablib Dataset representation of the Records.
        """
        import tablib

        # Create a new Tablib Dataset.
        data = tablib.Dataset()

        # If the RecordCollection is empty, just return the empty set
        # Check number of rows by typecasting to list
        if len(list(self)) == 0:
            return data

        data.headers = self.meta
        for row in self.all():
            row = reduce_datetimes(row.values())
            data.append(row)

        return data

    def all(self, asdict=False):
        """
        Returns a list of all rows for the RecordCollection. If they haven't
        been fetched yet, consume the iterator and cache the results.
        @params:
            `asdict`: Whether convert the records to dicts.
                - `False`: don't convert, keep `Record`
                - `True`:  convert to plain dict
                - `ordered`: convert to `OrderedDict`
        """

        # By calling list it calls the __iter__ method
        if not asdict:
            return list(self)
        if asdict is True:
            return [r.as_dict() for r in self]
        return [r.as_dict(True) for r in self]

    def first(self, default=None):
        """
        Returns a single record for the RecordCollection, or `default`. If
        `default` is an instance or subclass of Exception, then raise it
        instead of returning it.
        """

        # Try to get a record, or return/raise default.
        try:
            record = self[0]
        except IndexError:
            if isinstance(default, Exception):
                raise default
            return default

        return record



"""Utilities for pymedoo"""


def always_list(x):
    """
    Always return a list
    """
    return [y.strip() for y in x.split(",")] if isinstance(x, str) else list(x)


def reduce_datetimes(row):
    """
    Receives a row, converts datetimes to strings.
    """
    row = list(row)

    for i, iterrow in enumerate(row):
        if hasattr(iterrow, "isoformat"):
            row[i] = iterrow.isoformat()
    return tuple(row)



"""Mysql database"""
import psycopg2
from ..base import Base
from ..dialect import Dialect


class DialectPgsql(Dialect):
    """Mysql dialect"""


class Pgsql(Base):
    """Mysql medoo wrapper"""

    def __init__(self, *args, **kwargs):
        super(Pgsql, self).__init__(*args, **kwargs)
        self.cursor = self.connection.cursor()
        self.dialect(DialectPgsql)

    def _connect(self, *args, **kwargs):
        arguments = {"port": 5432}
        arguments.update(kwargs)
        return psycopg2.connect(**arguments)



"""Oracle database"""
import cx_Oracle
from ..base import Base
from ..dialect import Dialect


class DialectOracle(Dialect):
    """Oracle dialect"""


class Oracle(Base):
    """Oracle medoo wrapper"""

    def __init__(self, *args, **kwargs):
        super(Oracle, self).__init__(*args, **kwargs)
        self.cursor = self.connection.cursor()
        self.dialect(DialectOracle)

    def _connect(self, *args, **kwargs):
        arguments = {
            # some default settings
        }
        arguments.update(kwargs)
        return cx_Oracle.connect(**arguments)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['medoo', 'medoo.database']

package_data = \
{'': ['*']}

extras_require = \
{'all': ['mysql-connector-python>=8,<9',
         'psycopg2>=2,<3',
         'pymssql>=2,<3',
         'cx-Oracle>=8,<9'],
 'mssql': ['pymssql>=2,<3'],
 'mysql': ['mysql-connector-python>=8,<9'],
 'oracle': ['cx-Oracle>=8,<9'],
 'postgresql': ['psycopg2>=2,<3']}

setup_kwargs = {
    'name': 'medoo',
    'version': '0.1.2',
    'description': 'A lightweight database framework for python',
    'long_description': '# pymedoo - A lightweight database framework for python\n\nit\'s inspired by [Medoo][1] for PHP and [Records][7] for python.\n\n[![Pypi][8]][15]\n[![Github][9]][16]\n[![Codacy][10]][17]\n[![Codacy coverage][11]][17]\n![Building][12]\n\n## Install\n\n```bash\npip install medoo\n```\n\n## Required packages for databases\n\n| Database | Package   | Install |\n|----------|---------  | --------|\n| sqlite   | [sqlite3][2]   | `pip install medoo[sqlite]` |\n| mysql    | ~~[pymysql][3]~~(dropped) Use [mysql.connector][13] instead. See [#6][14] | `pip install medoo[mysql]` |\n| pgsql    | [psycopg2][4]  | `pip install medoo[pgsql]` |\n| mssql    | [pymssql][5]   | `pip install medoo[mssql]` |\n| oracle   | [cx_Oracle][6] | `pip install medoo[oracle]` |\n\nInstall forr all supported databases:\n\n```bash\npip install medoo[all]\n```\n\n## Get started\n\n### SELECT\n\n```python\nfrom medoo import Medoo\n\n# For other arguments, please refer to the original connect function of each client.\nme = Medoo(dbtype = \'sqlite\', database = \'file:///path/to/test.sqlite\')\n\n# SELECT * FROM "Customers"\nrs = me.select(\'Customers\')\n\nprint(rs.export(\'csv\', delimiter = \'\\t\'))\n```\n\n|CustomerID|CustomerName|ContactName|Address|City|PostalCode|Country|\n|-|-|-|-|-|-|-|\n|1|Alfreds Futterkiste|Maria Anders|Obere Str. 57|Berlin|12209|Germany|\n|2|Ana Trujillo Emparedados y helados|Ana Trujillo|Avda. de la Constitución 2222|México D.F.|5021|Mexico|\n|3|Antonio Moreno Taquería|Antonio Moreno|Mataderos 2312|México D.F.|5023|Mexico|\n|4|Around the Horn|Thomas Hardy|120 Hanover Sq.|London|WA1 1DP|UK|\n|5|Berglunds snabbköp|Christina Berglund|Berguvsvägen 8|Luleå|S-958 22|Sweden|\n\n```python\n# SELECT "CustomerID","CustomerName" FROM "Customers"\nme.select(\'Customers\', \'CustomerID, CustomerName\')\nme.select(\'Customers\', [\'CustomerID\', \'CustomerName\'])\n\n# SELECT "C"."CustomerID" AS "CustomerID","C"."CustomerName" AS "name" FROM "Customers" AS "C"\nme.select(\'Customers(C)\', [\'C.CustomerID(id)\', \'C.CustomerName(name)\'])\n\n# SELECT DISTINCT "Country" FROM "Customers"\nme.select(\'Customers\', \'Country\', distinct = True)\n\n# SELECT COUNT("CustomerID") FROM "Customers"\nme.select(\'Customers\', \'CustomerID|COUNT\')\n\n# SELECT COUNT(DISTINCT "CustomerID") AS "c" FROM "Customers"\nme.select(\'Customers\', \'CustomerID|.COUNT(c)\')\n\n# SELECT "CustomerID"+1 FROM "Customers"\nfrom medoo import Field, Raw\nme.select(\'Customers\', Field(\'CustomerID\')+1)\n\n# SELECT \'Name: \' || CustomerName AS name FROM "Customers"\nrs = me.select(\'Customers\', Raw("\'Name: \' || CustomerName AS name"))\nfor r in rs: print(r.name)\n```\n\n```\nName: Alfreds Futterkiste\nName: Ana Trujillo Emparedados y helados\nName: Antonio Moreno Taquería\nName: Around the Horn\nName: Berglunds snabbköp\n```\n\n### WHERE\n\n#### Single condition\n\n```python\n# SELECT * FROM "Customers" WHERE "CustomerID" = 1\nme.select(\'Customers\', where = {\'CustomerID\': 1})\n\n# SELECT * FROM "Customers" WHERE "CustomerID" < 3\nme.select(\'Customers\', where = {\'CustomerID[<]\': 3})\n\n# SELECT * FROM "Customers" WHERE "CustomerID" IN (1,2,3)\nme.select(\'Customers\', where = {\'CustomerID\': (1,2,3)})\n\n# SELECT * FROM "Customers" WHERE "CustomerName" LIKE \'%b%\' OR "CustomerName" LIKE \'%c%\'\nme.select(\'Customers\', where = {\'CustomerName[~]\': (\'a\', \'b\')})\n\n# SELECT * FROM "Customers" WHERE "CustomerID" BETWEEN 1 AND 3\nme.select(\'Customers\', where = {\'CustomerID[<>]\': (1,3)})\n\n# SELECT * FROM "Customers" WHERE NOT "CustomerID" BETWEEN 1 AND 3\nme.select(\'Customers\', where = {\'!CustomerID[<>]\': (1,3)})\n\n# SELECT * FROM "Customers" WHERE "CustomerID" IS NULL\nme.select(\'Customers\', where = {\'CustomerID[is]\': None}) # where = {\'id[==]\': None}\n\n# SELECT * FROM "Customers" WHERE INSTR("CustomerName", \'Antonio\')\nme.select(\'Customers\', where = {Raw(\'INSTR("CustomerName", \\\'Antonio\\\')\'):None})\n```\n\n#### Compond\n\n```python\n# SELECT * FROM "Customers" WHERE "CustomerID" IN (1,2,3) AND "CustomerName" LIKE \'%b%\'\nme.select(\'Customers\', where = {\n    \'CustomerID\': (1,2,3),\n    \'CustomerName[~]\': \'b\'\n})\n# SELECT * FROM "Customers"\n# WHERE ("CustomerID" IN (1,2,3) AND "CustomerName" LIKE \'%b%\') AND\n# ("CustomerName" = \'cd\' OR "CustomerID" = 2) AND\n# ("CustomerID" < 3 AND NOT "CustomerName" = \'bc\')\nme.select(\'Customers\', where = {\n    \'AND\': {\n        \'CustomerID\': (1,2,3),\n        \'CustomerName[~]\': \'b\'\n    },\n    \'OR\': {\n        \'CustomerName\': \'cd\',\n        \'CustomerID\': 2\n    },\n    # you can use comment to distinguish multiple ANDs and ORs\n    \'AND #2\': {\n        \'CustomerID[<]\': 3,\n        \'!CustomerName\': \'bc\'\n    }\n})\n```\n\n#### Modifier\n\n```python\n# SELECT * FROM "Customers" ORDER BY "CustomerID" DESC, "CustomerName" ASC LIMIT 2 OFFSET 1\n# MSSQL:\n# SELECT * FROM "Customers" ORDER BY "CustomerID" DESC, "CustomerName" ASC\n# OFFSET 1 ROWS FETCH NEXT 2 ROWS ONLY\nme.select(\'Customers\', where = {\n    \'ORDER\': {\'CustomerID\': \'desc\', \'CustomerName\': \'asc\'},\n    \'LIMIT\': (2, 1)\n})\n\n# SELECT COUNT("CustomerID") AS "c","CustomerName" FROM "Customers" GROUP BY "Country" HAVING "CustomerID" > 1\nme.select(\'Customers\', \'CustomerID|count(c), CustomerName\', where = {\n    \'GROUP\': \'Country\',\n    \'HAVING\': {\'CustomerID[>]\': 1}\n})\n```\n\n### Using subquery\n\n```python\nprint(me.select(\'Orders\').export(\'csv\', delimiter = \'\\t\'))\n```\n\n|OrderID|CustomerID|OrderDate|\n|-|-|-|\n|10308|2|1996-09-18|\n|10309|37|1996-09-19|\n|10310|77|1996-09-20|\n\n```python\n# SELECT * FROM "Customers" AS "C",(SELECT "CustomerID" FROM "Orders") AS "O"\n#   WHERE "C"."CustomerID" = "O"."CustomerID"\nme.select([\n    \'Customers(C)\', # the first table\n    me.builder.select(\'Orders\', \'CustomerID\', sub = \'O\')\n], where = {\n    \'C.CustomerID\': Field(\'O.CustomerID\')\n})\n\n# SELECT * FROM "Customers" WHERE "CustomerID" IN (SELECT "CustomerID" FROM "Orders")\nme.select(\'Customers\', where = {\n    \'CustomerID\': me.builder.select(\'Orders\', \'CustomerID\')\n})\n```\n\n### JOIN\n\n```python\n# SELECT "O"."OrderID","C"."CustomerName","O"."OrderDate" FROM "Orders" AS "O"\n#   INNER JOIN "Customers" AS "C" ON "C"."CustomerID"="O"."CustomerID"\nme.select(\'Orders(O)\', \'O.OrderID,C.CustomerName,O.OrderDate\', join = {\n    \'Customers(C)\': \'CustomerID\'\n})\n\n# equivalent to\nme.select(\'Orders(O)\', \'O.OrderID,C.CustomerName,O.OrderDate\', join = {\n    \'[><]Customers(C)\': \'CustomerID\'\n})\n# [>] LEFT JOIN, [<] RIGHT JOIN [<>] FULL OUTER JOIN\n\n# Join on multiple columns (same in different tables)\n# join = { \'[><]Customers(C)\': [\'CustomerID\', \'OtherColumn\'] }\n\n# Join on different columns: JOIN "Customers" AS "C" ON "C"."CustomerID"="O"."OtherID"\n# join = { \'[><]Customers(C)\': {\'CustomerID\', \'OtherID\'} }\n\n# You can join multiple tables, use OrderedDict if you want to keep the order.\n```\n\n### UNION\n\n```python\n# SELECT "CustomerID" FROM "Customers" UNION SELECT "CustomerID" FROM "Orders"\nme.union(\n    me.builder.select(\'Customers\', \'CustomerID\'),\n    me.builder.select(\'Orders\', \'CustomerID\')\n)\n\n# SELECT "CustomerID" FROM "Customers" UNION ALL SELECT "CustomerID" FROM "Orders"\nme.union(\n    me.builder.select(\'Customers\', \'CustomerID\'),\n    me.builder.select(\'Orders\', \'CustomerID\', sub = True)\n)\n```\n\n### Records\n\n`Medoo.select` and `Medoo.union` return a collection of records, which is basically a generator, but you can still get items from it, as it will consume the generate if necessary. The idea is borrowed from [Records][7].\n\n```python\nrecords = me.select(\'Customers\', \'CustomerID(id)\')\nrecord  = records.first() # <Record {\'id\': 1}>\n\n# equivalent to\nrecord  = records[0]\n\n# you may also select other rows: records[1], records[2]\n# or return all rows:\nprint(records.all())\n\n# you can also export the records\n# this is the courtesy from tablib (https://github.com/kennethreitz/tablib)\n# check the kwargs with its documentation\nprint(records.export(\'csv\', delimiter = \'\\t\'))\n\n# You can also apply tablib\'s other function on the data:\n# records.tldata.<function>(<args>)\n\n# to get the value of each field from a record:\nprint(record[0]) # 1\nprint(record[\'id\']) # 1\nprint(record.id) # 1\nprint(record.as_dict()) # {\'id\': 1}\n```\n\n### INSERT\n\n```python\n# INSERT INTO "Orders" ("OrderID","CustomerID","OrderDate") VALUES (1,2,\'1999-09-09\'),(2,8,\'2001-10-12\')\nme.insert(\n    \'Orders\', # table\n    \'OrderID, CustomerID, OrderDate\', # fields\n    (1,2,\'1999-09-09\'), # values\n    (2,8,\'2001-10-12\')\n    # ...\n)\n# get the last insert row id\nprint(me.id()) # 5\n\n# INSERT INTO "Orders" ("OrderID","CustomerID","OrderDate") VALUES (1,2,\'1999-09-09\'),(2,8,\'2001-10,12\')\nme.insert(\n    \'Orders\', # table\n    {\'OrderID\': 1, \'CustomerID\': 2, \'OrderDate\': \'1999-09-09\'}, # fields with the first value\n    (2,8,\'2001-10-12\')\n    # ...\n)\nme.insert(\n    \'Orders\', # table\n    {\'OrderID\': 1, \'CustomerID\': 2, \'OrderDate\': \'1999-09-09\'}, # fields with the first value\n    {\'OrderID\': 2, \'CustomerID\': 8, \'OrderDate\': \'2001-10-12\'}  # specify the fields as well\n    # ...\n)\n# Or if your values have all the fields\n# INSERT INTO "Orders" VALUES (1,2,\'1999-09-09\'),(2,8,\'2001-10-12\')\nme.insert(\n    \'Orders\', # table\n    (1,2,\'1999-09-09\')\n    (2,8,\'2001-10-12\')\n    # ...\n)\n\n# You may hold the changes until all data inserted\nme.insert(..., commit = False)\nme.insert(..., commit = False)\nme.insert(..., commit = False)\nme.insert(..., commit = False)\nme.commit()\n# This applies with UPDATE and DELETE as well.\n```\n\n### UPDATE\n\n```python\n# UPDATE "Orders" SET "CustomerID"=10 WHERE "OrderID" = 2\nme.update(\n    \'Orders\', # table\n    data  = {\'CustomerID\': 10},\n    where = {\'OrderID\': 2}\n)\n# UPDATE "Orders" SET "CustomerID"="CustomerID"+1 WHERE "OrderID" = 2\nme.update(\n    \'Orders\', # table\n    data  = {\'CustomerID[+]\': 1},\n    where = {\'OrderID\': 2}\n)\n```\n\n### DELETE\n\n```python\n# DELETE FROM "Orders" WHERE "OrderID" = 2\nme.delete(\'Orders\', where = {\'OrderID\': 2})\n```\n\n### Other functions of `Medoo`\n\n```python\n# Fetch a single value\nme.get(\'Customers\', \'CustomerID\', where = {\'CustomerName\': \'Around the Horn\'}) # == 1\n\n# Check if a record exists\nme.has(\'Customers\', where = {\'CustomerID\': 10}) # == False\n\n# Return the last query\nme.last() # SELECT * FROM "Customers" WHERE "CustomerID" = 10\n\n# Show all the queries bound with `me`\n\n# You have to passing `logging = True` to `Medoo(..., logging = True)`\nme.log()\n\n# Return the errors\nme.error()\n\n# Submit an SQL query\nme.query(sql, commit = True)\n```\n\n### Extending `pymedoo`\n\n`pymedoo` is highly extendable, including the operators in `WHERE` conditions and `UPDATE SET` clause, `JOIN` operators, and some functions such as how to quote the table names, field names and values. All of these have been defined with `Dialect` class, what you need to do is just extend this class and specify it to the `Medoo` instance.\nFor example, let\'s define a case-insensitive `LIKE` operator using a shortcut `~~`:\n\n```python\nfrom medoo import Medoo, Dialect\n\nclass MyDialect(Dialect):\n    OPERATOR_MAP = {\n        \'~~\': \'ilike\'\n    }\n\n    @classmethod\n    def ilike(klass, field, value):\n        # support single value\n        if not isinstance(value, list):\n            value = [value]\n\n        terms = [\n            "UPPER({}) LIKE UPPER({})".format(field, klass.value(v)) # quote the value\n            for v in value\n        ]\n        # use OR to connect\n        return \' OR \'.join(terms)\n\n# tell medoo to use this dialect\nme = Medoo(...)\nme.dialect(MyDialect)\n\n# SELECT * FROM "Customers" WHERE UPPER("CustomerName") LIKE UPPER(\'%an%\')\nrecords = me.select(\'Customers\', where = {\n    \'CustomerName[~~]\': \'%an%\'\n})\nprint(records.export(\'csv\', delimiter = \'\\t\'))\n```\n\n|CustomerID|CustomerName|ContactName|Address|City|PostalCode|Country|\n|-|-|-|-|-|-|-|\n|2|Ana Trujillo Emparedados y helados|Ana Trujillo|Avda. de la Constitución 2222|México D.F.|5021|Mexico|\n|3|Antonio Moreno Taquería|Antonio Moreno|Mataderos 2312|México D.F.|5023|Mexico|\n\n[1]: https://medoo.in/\n[2]: https://docs.python.org/2/library/sqlite3.html\n[3]: https://github.com/PyMySQL/PyMySQL\n[4]: http://initd.org/psycopg/docs/\n[5]: http://www.pymssql.org/en/stable/\n[6]: https://oracle.github.io/python-cx_Oracle/\n[7]: https://github.com/kennethreitz/records\n[8]: https://img.shields.io/pypi/v/medoo.svg?style=flat-square\n[9]: https://img.shields.io/github/tag/pwwang/pymedoo.svg?style=flat-square\n[10]: https://img.shields.io/codacy/grade/c049a2bf5ca84484b885cfbfedc27eab.svg?style=flat-square\n[11]: https://img.shields.io/codacy/coverage/c049a2bf5ca84484b885cfbfedc27eab.svg?style=flat-square\n[12]: https://img.shields.io/github/workflow/status/pwwang/pymedoo/Build%20and%20Deploy?style=flat-square\n[13]: https://dev.mysql.com/doc/connector-python/en/\n[14]: https://github.com/pwwang/pymedoo/issues/6\n[15]: https://pypi.org/project/pymedoo/\n[16]: https://github.com/pwwang/pymedoo\n[17]: https://app.codacy.com/gh/pwwang/pymedoo\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pwwang/pymedoo',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Dialect for different databases"""
from .exception import WhereParseError, AnyAllSomeParseError

# import builder


class Dialect:
    """Dialect class"""

    OPERATOR_MAP = {
        "=": "eq",
        "~": "like",
        "!": "ne",
        "!=": "ne",
        "<>": "between",
        "==": "is_",
        "is": "is_",
    }

    UPDATE_MAP = {"=": "up_eq"}

    JOIN_MAP = {
        ">": "LEFT JOIN",
        "<": "RIGHT JOIN",
        "<>": "FULL OUTER JOIN",
        "><": "INNER JOIN",
    }

    @staticmethod
    def quote(item):
        """How to quote values"""
        if isinstance(item, str):
            if item == "*":
                return item
            return '"%s"' % item.replace('"', '""')
        # TODO: possible injection
        return str(item)

    @staticmethod
    def value(item):
        """How is VALUE being quoted"""
        if isinstance(item, str):
            return "'%s'" % item.replace("'", "''")
        # TODO: possible injection
        return str(item)

    @classmethod
    def limit(cls, limit, offset=None):
        """How is LIMIT being interpreted"""
        fmt = "LIMIT {limit}"
        if offset:
            fmt += " OFFSET {offset}"
        return fmt.format(limit=limit, offset=offset)

    @classmethod
    def up_eq(cls, field, value):
        """Equal (assignment) in UPDATE clause"""
        return "{}={}".format(field, cls.value(value))

    @classmethod
    def is_(cls, field, value):
        """How is IS being interpreted"""
        if value is not None:
            raise WhereParseError("IS is only used to tell NULL (None)")
        return "{} IS NULL".format(field)

    @classmethod
    def eq(cls, field, value):  # pylint: disable=invalid-name
        """How is EQ being interpreted"""
        from . import builder

        if isinstance(value, (tuple, list)) and len(value) == 1:
            value = value[0]

        if isinstance(value, (tuple, list)):
            return "{} IN ({})".format(
                field, ",".join([cls.value(v) for v in value])
            )
        if isinstance(value, builder.Builder):  # subquery
            return "{} IN ({})".format(field, cls.value(value))
        return "{} = {}".format(field, cls.value(value))

    @classmethod
    def like(cls, field, value, addperct=True):
        """How is LIKE being interpreted"""
        if isinstance(value, (tuple, list)) and len(value) == 1:
            value = value[0]

        if isinstance(value, (tuple, list)):
            if addperct:
                value = [
                    "%{}%".format(v)
                    if not v.startswith("%") and not v.endswith("%")
                    else v
                    for v in value
                    if isinstance(v, str)
                ]
            return "({})".format(
                " OR ".join(
                    ["{} LIKE {}".format(field, cls.value(v)) for v in value]
                )
            )

        if (
            addperct
            and isinstance(value, str)
            and not value.startswith("%")
            and not value.endswith("%")
        ):
            value = "%{}%".format(value)
        return "{} LIKE {}".format(field, cls.value(value))

    @classmethod
    def ne(cls, field, value):  # pylint:disable=invalid-name
        """How is NE being interpreted"""
        from . import builder

        if isinstance(value, (tuple, list)) and len(value) == 1:
            value = value[0]

        if isinstance(value, (tuple, list)):
            return "{} NOT IN ({})".format(
                field, ",".join([cls.value(v) for v in value])
            )
        if isinstance(value, builder.Builder):
            return "{} NOT IN ({})".format(field, cls.value(value))
        return "{} <> {}".format(field, cls.value(value))

    @classmethod
    def between(cls, field, value):
        """How is BETWEEN being interpreted"""
        if not isinstance(value, (tuple, list)) or len(value) != 2:
            raise WhereParseError(
                "BETWEEN value should a tuple or " "list with 2 elements."
            )
        return "{} BETWEEN {} AND {}".format(
            field, cls.value(value[0]), cls.value(value[1])
        )

    @classmethod
    def _default(cls, oprt, field, value):
        from . import builder

        if (
            oprt.lower().endswith("any")
            or oprt.lower().endswith("all")
            or oprt.lower().endswith("some")
        ):
            oprt = oprt.upper()
            if not isinstance(value, builder.Term):
                raise AnyAllSomeParseError(
                    "Require a subquery for " "ALL/ANY/SOME statement."
                )
            value = "({})".format(value)
        else:
            value = cls.value(value)
        return "{} {} {}".format(field, oprt, value)

    @classmethod
    def _up_default(cls, oprt, field, value):
        value = cls.value(value)
        return "{0}={0}{1}{2}".format(field, oprt, value)

    @classmethod
    def _join_default(cls, jointype):
        return jointype

    @classmethod
    def _operator(cls, oprt, field, value):
        oprt = oprt or "="
        oprtmap = cls.OPERATOR_MAP
        if oprt in oprtmap:
            oprt = oprtmap[oprt]
        else:
            sup = super(cls, cls)
            if hasattr(sup, "OPERATOR_MAP"):
                oprt = sup.OPERATOR_MAP.get(oprt, oprt)

        if hasattr(cls, oprt):
            return getattr(cls, oprt)(field, value)
        return cls._default(oprt, field, value)

    @classmethod
    def _update(cls, oprt, field, value):
        oprt = oprt or "="
        upmap = cls.UPDATE_MAP
        if oprt in upmap:
            oprt = upmap[oprt]
        else:
            sup = super(cls, cls)
            if hasattr(sup, "UPDATE_MAP"):
                oprt = sup.UPDATE_MAP.get(oprt, oprt)

        if hasattr(cls, oprt):
            return getattr(cls, oprt)(field, value)
        return cls._up_default(oprt, field, value)

    @classmethod
    def _join(cls, jointype=None):
        jointype = jointype or "><"
        joinmap = cls.JOIN_MAP
        if jointype in joinmap:
            jointype = joinmap[jointype]
        else:
            sup = super(cls, cls)
            if hasattr(sup, "JOIN_MAP"):
                jointype = sup.JOIN_MAP.get(jointype, jointype)

        if hasattr(cls, jointype):
            return getattr(cls, jointype)()
        return cls._join_default(jointype)
